#!/usr/sbin/perl -w
#
# GapSim.pl v1.0a
# Last modified April 29, 2004 11:53
# (c) Olaf R.P. Bininda-Emonds
#
# Creates random distributions of events based on characteristics of a Normal Table
#
# Usage: perl GapSim.pl -f<filename> [-n<replicates>] [-v] [-h]
#   options: -f<filename> = name of tab-delimited datafile in format species, total number of events,
#                           total number of stages, number of stages used, number of simultaneous events
#            -n<replicates> = number of random distributions to be created
#                             (default = 1000)
#            -v = verbose output
#            -h = print this message and quit

# Set defaults
	my $inFile;
	my $maxReps = 1000;
	my $verbose = 0;

for (my $i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i] =~ /-f([\w+\.?\w+?]+)/)
		{
		$inFile = $1;
		}
	elsif ($ARGV[$i] =~ /-n([\d]+)/)
		{
		$maxRep = $1;
		}
	elsif ($ARGV[$i] =~ /-v/)
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: GapSim.pl -f<filename> [-n<replicates>] [-v] [-h]\n";
		print "Options: -f<filename> = name of tab-delimited datafile in format species, total number of events,\n";
		print "                        total number of stages, number of stages used, number of simultaneous events\n";
		print "         -n<replicates> = number of random distributions to be created\n";
		print "                          (default = 1000)\n";
		print "         -v = verbose output\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "Usage: GapSim.pl -f<filename> [-n<replicates>] [-v] [-h]\n";
		exit(1); 
		}
	}

die "Must specify filename containing Normal Table data\n" if (not defined $inFile);
	($outFile) = split(/\./,$inFile,0);
	$outFile .= ".gap.txt";

setLineBreak($inFile);
open (fIN,"<$inFile");
open (fOUT,">$outFile");
	print fOUT "Results based on $maxRep replicates\n";
	print fOUT "Species\tEvents\tStages\tUsed Stages\tMean SimUsed\tSD SimUsed\tZ SimUsed\tSimul EPs\tMin Simul EPs\tMean SimSimul\tSD SimSimul\tZ SimSimul\n";

while (<fIN>)
	{
	$UsedMean = $UsedVar = $SimulMean = $SimulVar = 0;
	chomp;
	if ($_)
		{
		($Species,$NumEvents,$NumStages,$UsedStages,$NumSimul) = split (/\t/);
		print "Species: $Species\tEvents: $NumEvents\tStages: $NumStages\tUsed Stages: $UsedStages\tSimul EPs:$NumSimul\n" if ($verbose);
		for ($rep = 1; $rep <= $maxRep; $rep++)
			{
			$SimUsed = $SimSimul = 0;
			@StageCount1 = @StageCount2 = 0;		# Clear event array for each replicate
			print "\tReplicate: $rep\t" if ($verbose);
			for (1..$NumEvents)		# Assign NumEvents events randomly to stages
				{
				$RndStage1 = int($NumStages*rand())+1;
				$RndStage2 = int($UsedStages*rand())+1;
				$StageCount1[$RndStage1]++;
				$StageCount2[$RndStage2]++;
				}
			for (1..$NumStages)		# Get information from simulated event array
				{
				if ($StageCount1[$_])	# Count number of stages used
					{
					$SimUsed++;
					}
				if ($StageCount2[$_])	# Count number of simultaneous events (relative to number of stages used)
					{
					$SimSimul += ($StageCount2[$_]*$StageCount2[$_] - $StageCount2[$_]) / 2;
					}
				}
			print "Stages Used: $SimUsed\tNo. of simul EPs: $SimSimul\n" if ($verbose);
			$UsedMean += $SimUsed;
			$UsedVar += $SimUsed * $SimUsed;
			$SimulMean += $SimSimul;
			$SimulVar += $SimSimul * $SimSimul;
			}
		$UsedMean /= $maxRep;
		print "\n\tMean used: $UsedMean\n" if ($verbose);
		$UsedVar = ($UsedVar - ($UsedMean)*($UsedMean)*$maxRep)/($maxRep-1);
			$UsedVar = 0.0001 if ($UsedVar == 0);
		$UsedSD = sqrt($UsedVar);
		print "\tSD used: $UsedSD\n" if ($verbose);
		$SimulMean /= $maxRep;
		print "\tMean simul: $SimulMean\n" if ($verbose);
		$SimulVar = ($SimulVar - ($SimulMean)*($SimulMean)*$maxRep)/($maxRep-1);
			$SimulVar = 0.0001 if ($SimulVar == 0);
		$SimulSD = sqrt($SimulVar);
		print "\tSD simul: $SimulSD\n" if ($verbose);
		$Zused = ($UsedStages - $UsedMean) / $UsedSD;
		$Zsimul = ($NumSimul - $SimulMean) / $SimulSD;
		
		$eventsBase = int($NumEvents / $UsedStages);	# Calculate minimum number of simultaneous EPs
			$stageXtra = $NumEvents - ($UsedStages * $eventsBase);
			$eventsXtra = $eventsBase + 1;
			$stageBase = $UsedStages - $stageXtra;
#			$EventCheck = ($eventsBase*$stageBase) + ($eventsXtra*$stageXtra);
		
			$minSimul = ($stageBase * ($eventsBase * $eventsBase - $eventsBase) / 2) + ($stageXtra * ($eventsXtra * $eventsXtra - $eventsXtra) / 2);
			print "\tMin Simul EPs: $minSimul\n" if ($verbose);
	
		printf fOUT "%s\t%4i\t%4i\t%4i\t%4.3f\t%4.3f\t%4.3f",$Species,$NumEvents,$NumStages,$UsedStages,$UsedMean,$UsedSD,$Zused;
		print "\tZ used: $Zused" if ($verbose);
		if (abs($Zused) >= 1.96)
			{
			print "*\n" if ($verbose);
			print fOUT "*";
			}
		else
			{
			print "\n" if ($verbose);
			}
		printf fOUT "\t%4i\t%4i\t%4.3f\t%4.3f\t%4.3f",$NumSimul,$minSimul,$SimulMean,$SimulSD,$Zsimul;
		print "\tZ simul: $Zsimul" if ($verbose);
		if (abs($Zsimul) >= 1.96)
			{
			print "*\n\n" if ($verbose);
			print fOUT "*";
			}
		else
			{
			print "\n\n" if ($verbose);
			}
		print fOUT "\n";
		}		
	}

close fIN;
close fOUT;

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

